#!/usr/bin/env python3
""" A readable and intuitive way to generate Regular Expressions """
__version__ = '1.9.2'

# Import this as a submodule
from . import generate
from .api import api
from .EZRegex import EZRegex
from .generate import generate_regex
from .invert import invert
from .python import *



import random
import re
from inspect import currentframe, getframeinfo

from rich import print as rprint
from rich.table import Table
from rich.text import Text
from test_generate import *

import ezregex.python as er
from ezregex import api
from ezregex.generate import *
from ezregex.invert import *
from ezregex.python import *

try:
    from Cope import debug
except ImportError:
    pass

from _groups import *
from _groups import _losers, _winners
from _regexs import *
from _regexs import _regexsLine


def runTests(singletons=True, _invert=True, replacement=True, _generate=True, testMethod=False, internal=False, operators=True, strictness=20, dontIncludePassed=True, invertBackend='re_parser', invert_tries=1):
    global ow
    if singletons:
        print("Testing EZRegex singletons...")
        # Test the various parameters of anyof()
        assert er.anyof('aiLmsux', split=True, chars=True)._compile(False) == "[aiLmsux]",                  f"Was supposed to be '[aiLmsux]', was actually '{er.anyof('aiLmsux', split=True, chars=True)._compile(False)}'"
        # assert er.anyof('aiLmsux', split=False, chars=True)._compile(False) == Error,                     f"Was supposed to be 'Error', was actually '{er.anyof('aiLmsux', split=False, chars=True)._compile(False)}'"
        assert er.anyof('aiLmsux', split=None, chars=True)._compile(False) == "[aiLmsux]",                  f"Was supposed to be '[aiLmsux]', was actually '{er.anyof('aiLmsux', split=None, chars=True)._compile(False)}'"
        assert er.anyof('aiLmsux', split=True, chars=False)._compile(False) == '(?:a|i|L|m|s|u|x)',         f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{er.anyof('aiLmsux', split=True, chars=False)._compile(False)}'"
        # assert er.anyof('aiLmsux', split=False, chars=False)._compile(False) == Error,                    f"Was supposed to be 'Error', was actually '{er.anyof('aiLmsux', split=False, chars=False)._compile(False)}'"
        assert er.anyof('aiLmsux', split=None, chars=False)._compile(False) == '(?:a|i|L|m|s|u|x)',         f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{er.anyof('aiLmsux', split=None, chars=False)._compile(False)}'"
        assert er.anyof('aiLmsux', split=True, chars=None)._compile(False) == '(?:a|i|L|m|s|u|x)',          f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{er.anyof('aiLmsux', split=True, chars=None)._compile(False)}'"
        # assert er.anyof('aiLmsux', split=False, chars=None)._compile(False) == Error,                     f"Was supposed to be 'Error', was actually '{er.anyof('aiLmsux', split=False, chars=None)._compile(False)}'"
        assert er.anyof('aiLmsux', split=None, chars=None)._compile(False) == "[aiLmsux]",                  f"Was supposed to be '[aiLmsux]', was actually '{er.anyof('aiLmsux', split=None, chars=None)._compile(False)}'"

        # assert er.anyof(*list('aiLmsux'), split=True, chars=True)._compile(False) == Error,               f"Was supposed to be 'Error', was actually '{er.anyof(*list('aiLmsux'), split=True, chars=True)._compile(False)}'"
        assert er.anyof(*list('aiLmsux'), split=False, chars=True)._compile(False) == "[aiLmsux]",          f"Was supposed to be '[aiLmsux]', was actually '{er.anyof(*list('aiLmsux'), split=False, chars=True)._compile(False)}'"
        assert er.anyof(*list('aiLmsux'), split=None, chars=True)._compile(False) == "[aiLmsux]",           f"Was supposed to be '[aiLmsux]', was actually '{er.anyof(*list('aiLmsux'), split=None, chars=True)._compile(False)}'"
        # assert er.anyof(*list('aiLmsux'), split=True, chars=False)._compile(False) == Error,              f"Was supposed to be 'Error', was actually '{er.anyof(*list('aiLmsux'), split=True, chars=False)._compile(False)}'"
        assert er.anyof(*list('aiLmsux'), split=False, chars=False)._compile(False) == '(?:a|i|L|m|s|u|x)', f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{er.anyof(*list('aiLmsux'), split=False, chars=False)._compile(False)}'"
        assert er.anyof(*list('aiLmsux'), split=None, chars=False)._compile(False) == '(?:a|i|L|m|s|u|x)',  f"Was supposed to be '(?:a|i|L|m|s|u|x)', was actually '{er.anyof(*list('aiLmsux'), split=None, chars=False)._compile(False)}'"
        # assert er.anyof(*list('aiLmsux'), split=True, chars=None)._compile(False) == Error,               f"Was supposed to be 'Error', was actually '{er.anyof(*list('aiLmsux'), split=True, chars=None)._compile(False)}'"
        assert er.anyof(*list('aiLmsux'), split=False, chars=None)._compile(False) == "[aiLmsux]",          f"Was supposed to be '[aiLmsux]', was actually '{er.anyof(*list('aiLmsux'), split=False, chars=None)._compile(False)}'"
        assert er.anyof(*list('aiLmsux'), split=None, chars=None)._compile(False) == "[aiLmsux]",           f"Was supposed to be '[aiLmsux]', was actually '{er.anyof(*list('aiLmsux'), split=None, chars=None)._compile(False)}'"

        assert er.anyCharExcept('abcd')._compile(False) == '[^abcd]'
        assert er.anyCharExcept(*list('abcd'))._compile(False) == '[^abcd]'

        # Test flags
        assert str(word + ASCII + stuff)      == r'(?a)\w+.+', fr"{word + ASCII + stuff}      != (?a)\w+.+"
        assert str(word) == r'\w+', f'{word}'
        assert str(word + DOTALL + stuff)     == r'(?s)\w+.+', fr"{word + DOTALL + stuff}     != (?s)\w+.+"
        assert str(word + IGNORECASE + stuff) == r'(?i)\w+.+', fr"{word + IGNORECASE + stuff} != (?i)\w+.+"
        assert str(word + LOCALE + stuff)     == r'(?L)\w+.+', fr"{word + LOCALE + stuff}     != (?L)\w+.+"
        assert str(word + MULTILINE + stuff)  == r'(?m)\w+.+', fr"{word + MULTILINE + stuff}  != (?m)\w+.+"
        # assert str(word + UNICODE + stuff)    == r'(?u)\w+.+', fr"{word + UNICODE + stuff}    != (?u)\w+.+"

        a = word + ow
        # b = stuff + UNICODE
        c = IGNORECASE + '9'
        assert a + c == word + ow + IGNORECASE + '9', f"{a + b + c} != {word + ow + IGNORECASE + '9'}"

        for cnt, r in enumerate(regexs):
            regex, match, dontMatch = r
            try:
                if match is not None:
                    for m in match:
                        assert m in regex, f"{regex} does not match '{m}' (approx. {__file__}, line {_regexsLine+cnt})"
                if dontMatch is not None:
                    for m in dontMatch:
                        assert m not in regex, f"{regex} DOES match '{m}' (approx. {__file__}, line {_regexsLine+cnt})"
            except Exception as err:
                print(regex)
                print(f'Error @ approx. {__file__}, line {_regexsLine+cnt}: \nregex = `{regex}`, match = `{match}`, dontMatch = `{dontMatch}`')
                raise err#.with_traceback(None)

        a = str(EZRegex(r'\s+', 'python'))
        b = str(EZRegex(raw(r'\s+'), 'python'))
        c = r'\s+'
        d = str(raw(r'\s+'))
        # e = str(whitespace + matchMax)
        assert a == b == c == d, f'\na: {a}\n b: {b}\n c: {c}\n d: {d}\n e: {e}'
        # assert (word + ow + anything + ':').test('word    d:', show=False)
        # assert not (word + ow + anything + ':').test('word', show=False)
        assert 'word    d:' in (word + ow + anything + ':')

        test = word + chunk
        test += word
        assert str(test) == str(word + chunk + word), f"{str(test)} != {str(word + chunk + word)}"
        assert test == word + chunk + word
        assert either('(' + word + ')', '.') == either(er.literal('(') + word() + er.literal(')'), '.'), f"{either('(' + word + ')', '.')} != {either(er.literal('(') + word() + er.literal(')'), '.')}"
        assert str(ifFollowedBy(word)) == r'(?=\w+)'

        #TODO: assert (word + ow + anything + ':') in 'word    d:'
        #TODO: assert (word + ow + anything + ':') not in 'word'

    if _invert:
        print("Testing Invert...")

        table = Table(title="invert tests", expand=False)
        table.add_column("Line", justify="center", style="grey37")#, max_width=2)
        table.add_column("Regex", justify="right", style="green")#, max_width=40)
        table.add_column("Inverse", justify="left", style="dim green")
        table.add_column("Success", justify="center")

        for cnt, r in enumerate(regexs + replacements):
            regex = r[0]

            try:
                for _ in range(strictness):
                    # -1 means return it even if it's bad
                    inv = invert(regex, backend=invertBackend, tries=invert_tries)
                    if inv not in regex or not dontIncludePassed:
                        table.add_row(str(_regexsLine+cnt), Text(regex.str()), '`' + inv + '`', Text('passed', style='blue') if inv in regex else Text('failed', style='red'))
            except (Exception, AssertionError) as err:
                print(f'Error @ approx. {__file__}, line {_regexsLine+cnt}: \nregex = `{regex}`')#, inv = `{inv}`')
                raise err#.with_traceback(None)
        if len(table.rows):
            rprint(table)

    if _generate:
        generate_tests()
        test_bb()
        test_new_parts()

        for w, l in zip(_winners, _losers):
            r = re.compile(generate_regex(w, l, 500, restarts=2))
            for i in w:
                assert r.search(i), f"`{r}` is not in `{i}`"
            for i in l:
                assert not r.search(i), f"`{r}` is in `{i}`"
            # print(r, 'is good')

    if replacement:
        for pattern, repl, s, ans in replacements:
            assert re.sub(str(pattern), str(repl), s) == ans, f'Replacing\n\t{pattern}\nwith\n\t{repl}\nin\n\t{s}\nyielded\n\t{re.sub(str(pattern), str(repl), s)}\nnot\n\t{ans}'

    if testMethod:
        # ow = optional(whitechunk)
        params = er.group(er.atLeastNone(er.ow + er.word + er.ow + er.optional(',') + er.ow))
        function = er.word + er.ow + '(' + params + ')'
        function.test('this should match func(param1, param2 ), foo(), and bar( foo,)')

        r = 'group 1' + ':' + ow + group('stuff') + ' | ' + 'group ' + number + ': ' + group('things') + ' | ' + 'named group "' + word + '": '  + named_group('foo', 'bar')
        s = 'random stuff! and then group 1: stuff | group 2: things | named group "foo": bar  \t oh and then more random stuff'
        r.test(s)

        s = 'word1 word2 word3'
        word.test(s)

        (word + whitechunk + group('func') + ':' + namedGroup('test', anyof('8', '7'))).test()

        # This is actually accurate, if you think about it.
        # ifFollowedBy(word).test("literal(hllow) + isExactly('thing')")# fails in _matchJSON()

        ('(' + +(anything + optional(group(comma))) + ')').test()# -- empty groups print as None

        group(+group(number) + group(anyof('98'))).test('999')

    if internal:
        # rprint((word + number)._matchJSON())
        # rprint((word + whitechunk + group('func') + ':' + namedGroup('test', anyof('8', '7')))._matchJSON())
        rprint(ifFollowedBy(word)._matchJSON())
        rprint(word._matchJSON())
        rprint(number._matchJSON('word'))
        r = 'group 1' + ':' + ow + group('stuff') + ' | ' + 'group ' + number + ': ' + group('things') + ' | ' + 'named group "' + word + '": '  + named_group('foo', 'bar')
        s = 'random stuff! and then group 1: stuff | group 2: things | named group "foo": bar  \t oh and then more random stuff'
        rprint(r._matchJSON(s))

    if operators:
        print('Testing operators...')
        # The ~ operator
        for r in random.choices(regexs, k=strictness):
            regex, match, dontMatch = r
            assert re.search(regex.str(), ~regex), f"Invertting with ~ failed to find {regex} in {~regex}"

        # Not sure why this doesn't work?...
        # try:
        #     not anything
        # except NotImplementedError:
        #     pass
        # else:
        #     assert False

        # This is weird...
        digit = er.digit

        # TODO: Finish this
        # assert digit * ... == matchMax(digit)

        # TODO: Debug this
        # assert digit | word == anyof(digit, word), f"{digit | word} != {anyof(digit, word)}"
        # assert 7 | digit == anyof(7, digit), f"{7 | digit} != {anyof(digit, 7)}"
        # assert re.search(str(digit | word | '1'), '1') == re.search(anyof(digit, word, 1).str(), '1')

        assert +digit == matchMax(digit)

        assert anything + word == anything << word
        assert anything + word == anything >> word

        assert digit[2, 3] == match_range(2, 3, digit)
        assert digit[2, ...] == digit[2,] == digit[2, None] == digit[2] == match_at_least(2, digit)
        assert digit[..., 2] == digit[0, 2] == digit[None, 2] == match_at_most(2, digit)
        assert digit[...] == digit[0, ...] == digit[None] == at_least_0(digit)
        assert digit[1, ...] == digit[1] == digit[1,] == digit[1, None] == at_least_1(digit)
        # expr[...:end_expr] is equivalent to ZeroOrMore(expr, stop_on=end_expr)
        # assert digit[...:'foo'] == digit[None:'foo'] == digit[,'foo'] ==

        test = digit
        test += 6
        assert test == digit + 6, f"{test} != {digit + 6}"

        test = digit
        test *= 3
        assert test == digit * 3

        assert 2 * digit == digit * 2

        assert digit % 'sldkj' is None, f'{digit % "sldkj"} != None'
        assert digit % '77sdsf88' == re.search('77sdsf88', digit.str())

        # Test f-strings
        # As of right now, these may work, but won't necissarily work
        # assert f'|{word}|' == '|' + word + '|'
        # assert f'|{word}|{number}|' == '|' + word + '|' + number + '|'
        # assert f"|{word}{number + ifFollowedBy(word)}|" == '|' + word + number + ifFollowedBy(word) + '|'
        # assert f'|{rgroup("g")}|{replace_group}|' == '|' + rgroup('g') + '|' + replace_entire + '|'

        # Replacement strings are, in fact, a thing
        # print(version_info)
        assert replace('|{g}|{1}|{0}|') == '|' + rgroup('g') + '|' + rgroup(1) + '|' + replace_entire + '|'
        assert replace("{group}this is am{group}mtest") == rgroup('group') + 'this is am' + rgroup('group') + "mtest"
        assert replace("this is {{ not a thing") == "this is { not a thing"
        assert replace("also not }} a thing") == "also not } a thing"
        assert replace("still }}not{{ a thing") == "still }not{ a thing"
        assert replace("also {{not}} a thing") == "also {not} a thing"
        assert replace("but {group} is and {1} is") == "but " + rgroup('group') + " is and " + rgroup(1) + " is"
        assert replace("{group}{g}") == rgroup('group') + rgroup('g')

        # no idea why this doesnt work.
        # assert (anything + word) * 3 == '.\w+' * 3, f"'{(anything + word) * 3}' != '{'.\w+'*3}'"

    print('All Tests Passed!')

# From 1-100, 1 is easy, 100 is hard
difficulty = 1
runTests(
    # These should remain on, for the GitHub automated tests
    singletons=False,
    _invert=True,
    replacement=False,
    operators=False,
    _generate=False,
    # These display for you to check that they look correct
    testMethod=False,
    # _api=False,
    # Settings
    strictness=difficulty,
    invert_tries=101-difficulty,
    dontIncludePassed=True,
    invertBackend='re_parser',
)



# This is the final say of what's supported
# Even though they can still import elements directly from the other modules, they'll fail if they're not here
dialects = {
    'python': {
        "beginning": '',
        "end": '',
        'flag_func': lambda final, flags: f'(?{flags}){final}',
        'escape_chars': b'()[]{}?*+-|^$\\.&~# \t\n\r\v\f',
    },
    'javascript': {
        "beginning": '/',
        "end": '/',
        'flag_func': lambda final, flags: final + flags,
        'escape_chars': b'()[]{}?*+-|^$\\.&~# \t\n\r\v\f',
    },
    'perl': {
        "beginning": '',
        "end": '',
        'flag_func': lambda final, flags: f'(?{flags}){final}',
        'escape_chars': b'()[]{}?*+-|^$\\.&~# \t\n\r\v\f',
    },

}



# pyright: reportArgumentType = false
import colorsys
import logging
import re
from copy import deepcopy
from functools import partial
from typing import Callable, List, Literal, LiteralString
from warnings import warn

from .api import api
from .generate import *
from .invert import invert
from ._dialects import dialects

# TODO: consider changing addFlags to "outer" or "end" or something
# TODO: Seriously consider removing the debug functions

class EZRegex:
    """ Represent parts of the Regex syntax. Should not be instantiated by the user directly."""

    def __init__(self,
                 definition:List[partial[str]]|str|"EZRegex"|partial[str]|list[str],
                 dialect: str,
                 sanatize:bool=True,
                 init:bool=True,
                 replacement:bool=False,
                 flags:str='',
        ):
        """
        The workhorse of the EZRegex library. This represents a regex pattern that can be combined
        with other EZRegexs and strings. Ideally, this should only be called internally, but it should
        still work from the user's end
        """

        if dialect not in dialects.keys():
            raise ValueError(f'Unsupported dialect `{dialect}` given. Supported dialects are: {list(dialects.keys())}')

        self.flags = flags
        # Parse params
        # Add flags if it's another EZRegex
        if isinstance(definition, EZRegex):
            self.flags = definition.flags
            if definition.dialect != dialect:
                raise ValueError('Cannot mix regex dialects')

        if isinstance(definition, (str, EZRegex)):
            definition = [str(definition)]
        elif not isinstance(definition, (list, tuple)):
            definition = [definition]

        self.sanatize = sanatize
        self.replacement = replacement
        # This allows strings in the list now, but they get converted later in this function
        self.funcList: list[str|partial[str]|Callable] = list(definition)
        # Just some psuedonymns
        self.example = self.invert = self.inverse
        self.dialect = dialect
        # The dict that has the values
        self._dialect = dialects[dialect]


        # The init parameter is not actually required, but it will make it more
        # efficient, so we don't have to check that the whole chain is callable
        if init:
            # Go through the chain (most likely of length 1) and parse any strings
            # This is for simplicity when defining all the members
            for i in range(len(self.funcList)):
                if isinstance(self.funcList[i], str):
                    # I *hate* how Python handles lambdas
                    stringBecauseHatred = deepcopy(self.funcList[i])
                    self.funcList[i] = lambda cur=...: cur + stringBecauseHatred
                elif not callable(self.funcList[i]) and self.funcList[i] is not None:
                    raise ValueError(f"Invalid type {type(self.funcList[i])} passed to EZRegex constructor")


    def _escape(self, pattern:str):
        """ This function was modified from the one in /usr/lib64/python3.12/re/__init__.py line 255 """
        _special_chars_map = {i: '\\' + chr(i) for i in self._dialect['escape_chars']}
        return pattern.translate(_special_chars_map)


    def _sanitizeInput(self, i, addFlags=False):
        """ Instead of rasising an error if passed a strange datatype, it now trys to cast it to a string """
        i = deepcopy(i)

        # Don't sanatize anything if this is a replacement string
        if self.replacement:
            return str(i)

        # If it's another chain, compile it
        if isinstance(i, EZRegex):
            return i._compile(addFlags=addFlags)
        # It's a string (so we need to escape it)
        elif isinstance(i, str):
            return self._escape(i)
        # A couple of singletons use bools and None as kwargs, just ignore them and move on
        elif i is None or isinstance(i, bool):
            return i
        # A couple singletons use ints as input, just cast it to a string and don't throw a warning
        elif isinstance(i, int):
            return str(i)
        # It's something we don't know, but try to cast it to a string anyway
        else:
            try:
                logging.warning(f"Type {type(i)} passed to EZRegex, auto-casting to a string. Special characters will will not be escaped.")
                return str(i)
            except:
                raise ValueError(f'Incorrect type {type(i)} given to EZRegex parameter: Must be string or another EZRegex chain.')

    def __call__(self, *args, **kwargs):
        """ This should be called by the user to specify the specific parameters of this instance i.e. anyof('a', 'b') """
        # If this is being called without parameters, just compile the chain.
        # If it's being called *with* parameters, then it better be a fundemental
        # member, otherwise that doesn't make any sense.
        if len(self.funcList) > 1:
            if not len(args) and not len(kwargs):
                return self._compile()
            else:
                raise ValueError("You're trying to pass parameters to a chain of expressions. That doesn't make any sense. Stop that.")

        # Sanatize the arguments
        args = list(map(self._sanitizeInput if self.sanatize else deepcopy, args))

        _kwargs = {}
        for key, val in kwargs.items():
            _kwargs[key] = self._sanitizeInput(val) if self.sanatize else deepcopy(val)

        return EZRegex(
            [partial(self.funcList[0], *args, **_kwargs)],
            dialect=self.dialect,
            init=False,
            sanatize=self.sanatize,
            replacement=self.replacement,
            flags=self.flags
        )

    # Magic Functions
    def __str__(self, addFlags=True):
        return self._compile(addFlags)

    def __repr__(self):
        return 'EZRegex("' + self._compile() + '")'

    def __eq__(self, thing):
        return self._sanitizeInput(thing, addFlags=True) == self._compile()

    def __mul__(self, amt):
        if amt is Ellipsis:
            return EZRegex(f'(?{self})*', self.dialect, sanatize=False)
        rtn = self
        # This isn't optimal, but it's unlikely anyone will use this with large numbers
        for i in range(amt-1):
            # This doesn't work
            # rtn += self
            # But this does??
            rtn = rtn + self
        return rtn

    def __rmul__(self, amt):
        return self * amt

    def __imul__(self, amt):
        return self * amt

    def __add__(self, thing):
        return EZRegex(self.funcList + [partial(lambda cur=...: cur + self._sanitizeInput(thing))],
            dialect=self.dialect,
            init=False,
            sanatize=self.sanatize or thing.sanatize if isinstance(thing, EZRegex) else self.sanatize,
            replacement=self.replacement or thing.replacement if isinstance(thing, EZRegex) else self.replacement,
            flags=(self.flags + thing.flags) if isinstance(thing, EZRegex) else self.flags
        )

    def __radd__(self, thing):
        return EZRegex([partial(lambda cur=...: self._sanitizeInput(thing) + cur)] + self.funcList,
            dialect=self.dialect,
            init=False,
            sanatize=self.sanatize or thing.sanatize if isinstance(thing, EZRegex) else self.sanatize,
            replacement=self.replacement or thing.replacement if isinstance(thing, EZRegex) else self.replacement,
            flags=(self.flags + thing.flags) if isinstance(thing, EZRegex) else self.flags
        )

    def __iadd__(self, thing):
        # return self + self._sanitizeInput(thing)
        return self + thing

    def __and__(self, thing):
        warn('The & operator is unstable still. Use each() instead.')
        return EZRegex(fr'(?={self}){thing}', self.dialect, sanatize=False)

    def __rand__(self, thing):
        warn('The & operator is unstable still. Use each() instead.')
        return EZRegex(fr'(?={thing}){self}', self.dialect, sanatize=False)

    # The shift operators just shadow the add operators
    def __lshift__(self, thing):
        return self.__add__(thing)

    def __rlshift__(self, thing):
        return self.__radd__(thing)

    def __ilshift__(self, thing):
        return self.__iadd__(thing)

    # I don't think right and left shifts should be any different, right?
    def __rshift__(self, thing):
        return self.__add__(thing)

    def __rrshift__(self, thing):
        return self.__radd__(thing)

    def __irshift__(self, thing):
        return self.__iadd__(thing)

    def __invert__(self):
        return self.invert()

    def __not__(self):
        raise NotImplementedError('The not operator is not implemented. What you probably want is one of anyExcept(), anyCharExcept(), ifNotProceededBy(), or ifNotPreceededBy()')

    def __pos__(self):
        comp = self._compile()
        return EZRegex(('' if not len(comp) else r'(?:' + comp + r')') + r'+', self.dialect, sanatize=False)

    def __ror__(self, thing):
        print('ror called')
        return EZRegex(f'(?:{self._sanitizeInput(thing)}|{self._compile()})', self.dialect, sanatize=False)

    def __or__(self, thing):
        warn('The or operator is unstable and likely to fail, if used more than twice. Use anyof() instead, for now.')
        return EZRegex(f'(?:{self._compile()}|{self._sanitizeInput(thing)})', self.dialect, sanatize=False)

    def __xor__(self, thing):
        return NotImplemented

    def __rxor__(self, thing):
        return NotImplemented

    def __mod__(self, other):
        """ I would prefer __rmod__(), but it doesn't work on strings, since __mod__() is already specified for string formmating. """
        # I don't need to check this, re will do it for me
        # if not isisntance(other, str):
            # raise TypeError(f"Can't search type {type(other)} ")
        return re.search(other, self._compile())

    def __hash__(self):
        if len(self.funcList) > 1:
            return hash(self._compile())
        # If we only have 1 function lined up, that means we haven't
        # been called at all. And that means we're one of the basic singletons,
        # because users aren't supposed to instantiate this class directly.
        # THAT means we can use this instance's pointer as a unique identifier.
        else:
            return hash(id(self))

    def __contains__(self, thing):
        assert isinstance(thing, str), "`in` statement can only be used with a string"
        return re.search(self._compile(), thing) is not None

    # I guess this isn't a thing? But it really should be.
    def __rcontains__(self, thing):
        assert isinstance(thing, str), "`in` statement can only be used with a string"
        return re.search(self._compile(), thing) is not None

    def __getitem__(self, args):
        # digit[2, 3]    # (2, 3)
        # digit[2, ...]  # (2, Ellipsis)
        # digit[2, None] # (2, None)
        # digit[2, ]     # (2,)
        # digit[..., 3]  # (Ellipsis, 3)
        # digit[None, 3] # (None, 3)
        # digit[...:3]   # slice(Ellipsis, 3, None)
        # digit[None:3]  # slice(None, 3, None)
        # digit[:3]      # slice(None, 3, None)
        # digit[:]       # slice(None, None, None)
        # digit[2]       # 2

        # assert digit[2, 3] == match_range(2, 3, digit)
        # assert digit[2, ...] == digit[2,] == digit[2, None] == digit[2] == match_at_least(2, digit)
        # # assert digit[..., 2] == digit[0, 2] == digit[None, 2] == match_at_most(2, digit)
        # assert digit[...] == digit[0, ...] == digit[None] == at_least_0(digit)
        # assert digit[1, ...] == digit[1] == digit[1,] == digit[1, None] == at_least_1(digit)

        if type(args) is slice:
            # expr[...:end_expr] is equivalent to ZeroOrMore(expr, stop_on=end_expr)
            # assert digit[...:'foo'] == digit[None:'foo'] == digit[,'foo'] ==
            pass
        elif type(args) is not tuple or len(args) == 1:
            if type(args) is tuple:
                args = args[0]
            if args is None or args is Ellipsis or args == 0:
                # at_least_0(self)
                return EZRegex(fr'(?:{self._compile()})*', self.dialect, sanatize=False)
            elif args == 1:
                # at_least_1(self)
                return EZRegex(fr'(?:{self._compile()})+', self.dialect, sanatize=False)
            else:
                # match_at_least(args, self)
                return EZRegex(fr'(?:{self._compile()}){{{args},}}', self.dialect, sanatize=False)
        else:
            start, end = args
            if start is None or start is Ellipsis:
                # match_at_most(2, self)
                return EZRegex(fr'(?:{self._compile()}){{0,{end}}}', self.dialect, sanatize=False)
            elif end is None or end is Ellipsis:
                if start == 0:
                    # at_least_0(self)
                    return EZRegex(fr'(?:{self._compile()})*', self.dialect, sanatize=False)
                elif start == 1:
                    # at_least_1(self)
                    return EZRegex(fr'(?:{self._compile()})+', self.dialect, sanatize=False)
                else:
                    # match_at_least(start, self)
                    return EZRegex(fr'(?:{self._compile()}){{{start},}}', self.dialect, sanatize=False)
            else:
                # match_range(start, end, self)
                return EZRegex(fr'(?:{self._compile()}){{{start},{end}}}', self.dialect, sanatize=False)

    def __reversed__(self):
        return self.inverse()

    def __rich__(self):
        return self._compile()

    def __pretty__(self):
        return self._compile()

    # Regular functions
    def _compile(self, addFlags=True):
        regex = ''
        for func in self.funcList:
            regex = func(cur=regex) # type: ignore

        # Add the flags
        if addFlags:
            regex = self._dialect['beginning'] + regex + self._dialect['end']
            if len(self.flags):
                regex = self._dialect['flag_func'](regex, self.flags)
        return regex

    def compile(self, addFlags=True):
        return re.compile(self._compile(addFlags))

    def str(self):
        return self.__str__()

    def debug(self):
        try:
            from Cope import debug
        except ModuleNotFoundError:
            print(f"Compiled ezregex string = {self}")
        else:
            debug(self, name='Compiled ezregex string', calls=2)
        return self

    def debugStr(self):
        return self.debug().str()

    def copy(self, addFlags=True):
        try:
            from clipboard import copy  # type: ignore
        except ImportError as err:
            raise ModuleNotFoundError('Please install the clipboard module in order to auto copy '
                                      'ezregex expressions (i.e. pip install clipboard)') from err
        else:
            copy(self._compile(addFlags=addFlags))

    def test(self, testString=None, show=True, context=True) -> bool:
        """ Tests the current regex expression to see if it's in @param testString.
            Returns the match objects (None if there was no match)"""
        from rich import print as rprint
        from rich.panel import Panel
        from rich.text import Text

        # json = self._matchJSON(testString=testString)
        json = api(self, test_string=testString)
        if not show:
            return bool(len(json['matches']))

        _cope = False
        if context:
            # Use the nice context function in the Cope library
            try:   from Cope import get_context, get_metadata
            except ImportError: pass
            else:  _cope = True

        st = Text()  # String
        gt = Text()  # Groups (all the group-related text)
        defaultColor = 'bold'
        textColor = ''

        st.append("Testing expression", style=defaultColor)
        # Add the context line
        if _cope:
            st.append(f' (from {get_context(get_metadata(2), False, True, True).strip()})', style=defaultColor)
        st.append(':\n', style=defaultColor)

        # The expression we're testing
        st.append(f'\t{json["regex"]}\n', style=textColor)
        st.append("for matches in:\n\t", style=defaultColor)

        # Add the main string
        for color, background, part in json['parts']:
            st.append(part, style=color if background is None else f'{color} on {background}')
        st.append('\n')

        # Add all the matches and groups
        for m in json['matches']:
            gt.append('Match = "')
            for color, background, part in m['match']['parts']:
                gt.append(part, style=color if background is None else f'{color} on {background}')
            gt.append('" ')
            gt.append(f"({m['match']['start']}:{m['match']['end']})", style='italic bright_black')
            gt.append('\n')
            if len(m['unnamed groups']):
                gt.append('Unnamed Groups:\n')
            for id, group in m['unnamed groups'].items():
                gt.append(f'\t{id}: "')
                gt.append(['string'], style=group['color'])
                gt.append('" ')
                gt.append(f"({group['start']}:{group['end']})", style='italic bright_black')
                gt.append('\n')
            if len(m['named groups']):
                gt.append('Named Groups:\n')
            for name, group in m['named groups'].items():
                gt.append(f'\t{name}: "')
                gt.append(group['string'], style=group['color'])
                gt.append('" ')
                gt.append(f"({group['start']}:{group['end']})", style='italic bright_black')
                gt.append('\n')
            gt.append('\n')

        # Assemble everything into a panel
        rprint(Panel(Text.assemble(*st, '\n', *gt),
            title='Testing Regex',
            subtitle=Text('Found\n', style='blue') if len(json['matches'])
                else Text('Not Found\n', style='red')))  #, border_style='dim green')

        return bool(len(json['matches']))

    def inverse(self, amt=1, **kwargs):
        """ "Inverts" the current Regex expression to give an example of a string it would match.
            Useful for debugging purposes. """
        return '\n'.join([invert(self._compile(), **kwargs) for _ in range(amt)])



# pyright: reportArgumentType = false
import re
from json import load

from ..base import load_base
from ..EZRegex import EZRegex

globals().update(load_base('javascript'))

# TODO: Use https://docs.python.org/3/library/string.html#string.Formatter instead
def replace(string, rtn_str=True):
    # Made with ezregex.org
    # Previous method
    # '{' + optional(anyCharExcept('{'), greedy=False) + group(+alphaNum) + '}' + optional(anyCharExcept('}'), greedy=False)
    # Current method
    # '{' + group(+alphaNum) + '}'
    r = r'\{((?:[A-Za-z0-9_])+)\}'
    # Convert them to something unique, then back, so we won't pick up things like {{this}}
    # {{this}} -> \0(this\0) -> {this}
    # instead of
    # {{this}} -> {this} -> \g<this>
    string = re.sub('{{', '\0\0(', string)
    string = re.sub('}}', '\0\0)', string)
    string = re.sub(r, r'\\g<\g<1>>', string)
    string = re.sub(r'\0\0\(', '{', string)
    string = re.sub(r'\0\0\)', '}', string)
    return string if rtn_str else EZRegex(string, 'javascript', sanatize=False, replacement=True)

rgroup = EZRegex(lambda num_or_name, cur=...: fr'{cur}\g<{num_or_name}>', 'javascript', replacement=True)
replace_entire = EZRegex(lambda cur=...: cur + r'\g<0>', 'javascript', replacement=True)



from .elements import *

matchMax = match_max
matchAtMost = atMost = at_most = match_at_most
matchNum = matchAmt = match_amt = amt = num = match_num
matchRange = match_range
matchMoreThan = match_greater_than = matchGreaterThan = moreThan = more_than = match_more_than
matchAtLeast = match_min = matchMin = atLeast = at_least = match_at_least
lineStartsWith = line_start = lineStart = line_starts_with
stringStartsWith = string_start = stringStart = string_starts_with
lineEndsWith = line_end = lineEnd = line_ends_with
stringEndsWith = string_end = stringEnd = string_ends_with
stuff      = chunk
whiteChunk = whitechunk
anychar    = anything
anyChar    = anything
char       = anything
alpha      = letter
alphanum   = alpha_num = alpha_num
white      = whitechunk

anyAmt = any_amt = zeroOrMore = zero_or_more = at_least_none
anyBetween = any_between
wordChar = word_char
hexDigit = hex
octDigit = oct_digit
newline = newLine = new_line
spaceOrTab = space_or_tab
carriageReturn = carriage_return
verticalTab = vertical_tab
formFeed = form_feed
dot = period
intOrFloat = int_or_float
notWhitespace = not_whitespace
notDigit = not_digit
notWord = not_word
anyof = any_of = anyOf = oneOf = one_of = any_of
anyExcept = any_except
anyCharExcept = any_char_except
printableAndSpace = printable_and_space
ifFollowedBy = if_followed_by = if_proceded_by
ifExists = if_exists
ifNotFollowedBy = if_not_followed_by = if_not_proceded_by
ifPrecededBy = if_preceded_by
ifNotPrecededBy = if_not_preceded_by
ifEnclosedWith = if_enclosed_by = ifEnclosedBy = if_enclosed_with
ifProcededBy = if_proceded_by
ifNotProcededBy = if_not_proceded_by
passiveGroup = passive_group
sameAs = same_as = earlierGroup = sameAsGroup = same_as_group = earlier_group
exactly = isExactly = is_exactly
oneOrNone = one_or_none = opt = optional
oneOrMore = one_or_more = atLeastOne = atLeast1 = at_least_1 = at_least_one
noneOrMore = none_or_more = atLeastNone = at_least_0 = atLeast0 = at_least_none
ascii = a = ASCII
dotall = s = DOTALL
ignorecase = i = ignoreCase = ignore_case = IGNORECASE
locale = L = LOCALE
multiline = m = MULTILINE
# Useful combinations
integer = signed
literallyAnything = literally_anything
wordBoundary = word_boundary
notWordBoundary = not_word_boundary

replaceGroup = replace_group = rgroup
replaceAll = replace_all = replaceEntire = replace_entire



import colorsys
import re

foreground_s = .75
foreground_v = 1.
background_v_bias = .5
background_s_bias = .9

# These functions comprise the color algorithm
def toHtml(r, g, b):
            return f'#{r:02x}{g:02x}{b:02x}'

def toRgb(html: str) -> tuple:
    hex_color = html.lstrip('#')
    rgb = tuple(int(hex_color[i:i+2], 16) for i in (0, 2, 4))
    return rgb

def generate_colors(amt, s:float=1, v:float=1, offset:int=0):
    """ Generate `amt` number of colors evenly spaced around the color wheel
        with a given saturation and value
    """
    amt += 1
    return [toHtml(*map(lambda c: round(c*255), colorsys.hsv_to_rgb(*((offset + ((1/amt) * (i + 1))) % 1.001, s, v)))) for i in range(amt-1)]

def furthest_colors(html, amt:int=5, v_bias:float=0, s_bias:float=0):
    """ Gets the `amt` number of colors evenly spaced around the color wheel from the given color
        `v_bias` and `s_bias` are between 0-1 and offset the colors
    """
    amt += 1
    h, s, v = colorsys.rgb_to_hsv(*map(lambda c: c/255, toRgb(html)))

    return [toHtml(*map(lambda c: round(c*255), colorsys.hsv_to_rgb(*((h + ((1/amt) * (i + 1))) % 1.001, (s+s_bias) % 1.001, (v+v_bias) % 1.001)))) for i in range(amt-1)]


def api(pattern, replacement_pattern=None, test_string=None, *, replacement_count=0, split_count=0):
    # Get an inverse, if nessicary
    if test_string is None:
        test_string = pattern.inverse()
    matches = list(re.finditer(pattern._compile(), test_string))
    found = bool(len(matches))

    json = {
        'regex': pattern._compile(),
        'string': test_string,
        'string HTML': ...,
        'parts': [],
        'matches': []
    }

    html_string = '<p><span style="color: white;">'
    parts = []
    globalCursor = 0
    allMatches = [m.span() for m in matches]
    # Map match spans to unique colors
    _colors = generate_colors(len(allMatches), s=foreground_s, v=foreground_v)
    matchColors = dict(zip(allMatches, _colors))

    for match in matches:
        allGroups = {match.span(i+1) for i in range(len(match.groups()))}
        namedGroups = {i: match.span(i) for i in match.groupdict().keys()}
        # TODO: have named groups show their name and number instead of just their name
        # namedGroups = {
        #     (cnt+1, ): match.span(cnt+1)
        #     for cnt, i in enumerate(match.groups())
        #     if i in match.groupdict().values()
        # }
        unnamedGroups = {
            cnt+1: match.span(cnt+1)
            for cnt, i in enumerate(match.groups())
            if i not in match.groupdict().values()
        }
        # Map group spans to unique colors
        # This gets equally spaced colors from the given color, so they're differentiable
        # and readable on a dark background
        colors = dict(zip(allGroups, furthest_colors(
            matchColors[match.span()],
            amt=len(allGroups),
            v_bias=background_v_bias,
            s_bias=background_s_bias
        )))
        cursor = match.span()[0]

        # First, get up until the match
        html_string += f'{test_string[globalCursor:cursor]}</span>'
        parts.append([None, None, test_string[globalCursor:cursor]])
        match_html = ''
        match_parts = []
        for g in sorted(allGroups, key=lambda x: x[0]):
            # This fixes the bug where overlapping groups get put in twice. By simply preventing
            # the cursor from moving backwards, we eliminate the latter (parent) group from being shown.
            if g[0] < cursor:
                continue

            # Print the match up until the group
            match_html += f'<span style="color: {matchColors[match.span()]};">{test_string[cursor:g[0]]}</span>'
            match_parts.append([matchColors[match.span()], None, test_string[cursor:g[0]]])

            # Print the group
            match_html += f'<span style="background-color: {colors[g]}; color: {matchColors[match.span()]};">{test_string[g[0]:g[1]]}</span>'
            match_parts.append([matchColors[match.span()], colors[g], test_string[g[0]:g[1]]])
            cursor = g[1]
        match_html += f'<span style="color: {matchColors[match.span()]};">{test_string[cursor:match.span()[1]]}</span>'
        match_parts.append([matchColors[match.span()], None, test_string[cursor:match.span()[1]]])
        globalCursor = match.span()[1]
        # Don't print after the group, cause there might be another match that covers it
        html_string += match_html
        parts += match_parts
        toSlice = lambda t: f'({t[0]}:{t[1]})'
        match_json = {
            'match': {
                'string': match.group(),
                'string HTML': match_html,
                'parts': match_parts,
                'end': match.end(),
                'start': match.start(),
                "color": matchColors[match.span()],
            },
            "unnamed groups":{},
            "named groups":{},
        }

        for num, span in unnamedGroups.items():
            match_json['unnamed groups'][num] = {
                'string': match.group(num) or '',
                'end': span[1],
                'start': span[0],
                "color": colors[span],
            }

        for name, span in namedGroups.items():
            match_json['named groups'][name] = {
                'string': match.group(name) or '',
                'end': span[1],
                'start': span[0],
                "color": colors[span],
            }
        json['matches'].append(match_json)

    # Don't forget to add any bit at the end that's not part of a match
    html_string += test_string[globalCursor:]
    parts.append([None, None, test_string[globalCursor:]])
    html_string += '</span></p>'
    json['string HTML'] = html_string
    json['parts'] = parts
    if replacement_pattern is not None:
        json['replaced'] = re.sub(pattern.str(), replacement_pattern.str(), test_string, replacement_count)
    else:
        json['replaced'] = None
    json['split'] = re.split(pattern.str(), test_string, split_count)
    return json



from ..EZRegex import EZRegex
from .elements import base


def load_base(dialect, exclude=[]) -> dict:
    rtn = {}
    for name, kwargs in base.items():
        rtn[name] = EZRegex(**kwargs, dialect=dialect)
    return rtn



# pyright: reportOperatorIssue = false
from re import escape
from string import digits
from sys import version_info

from ezregex import EZRegex


def _match_range_func(min, max, input, greedy=True, possessive=False, cur=...):
    assert not ((not greedy) and possessive), 'Match Range can\'t be non-greedy AND possessive at the same time'
    s = cur
    if len(input):
        s += r'(?:' + input + r')'
    s += r'{' + str(min) + r',' + str(max) + r'}'
    if not greedy:
        s += r'?'
    if possessive:
        if version_info < (3, 11):
            raise Exception('Possessive qualifiers require at least Python3.11 ')
        s += r'+'
    return s

def _at_least_one_func(input, greedy=True, possessive=False, cur=...):
    assert not ((not greedy) and possessive), 'At Least One can\'t be non-greedy AND possessive at the same time'
    s = cur
    if len(input) > 1:
        s += fr'(?:{input})+'
    else:
        if len(input) == 1:
            s += fr'{input}+'
    if not greedy:
        s += '?'
    if possessive:
        if version_info < (3, 11):
            raise Exception('Possessive qualifiers require at least Python3.11 ')
        s += '+'
    return s

def _at_least_none_func(input, greedy=True, possessive=False, cur=...):
    assert not ((not greedy) and possessive), 'At Least None can\'t be non-greedy AND possessive at the same time'
    s = cur
    if len(input) > 1:
        s += fr'(?:{input})*'
    else:
        if len(input) == 1:
            s += fr'{input}*'
    if not greedy:
        s += '?'
    if possessive:
        if version_info < (3, 11):
            raise Exception('Possessive qualifiers require at least Python3.11 ')
        s += '+'
    return s

def _optional_func(input, greedy=True, possessive=False, cur=...):
    assert not ((not greedy) and possessive), 'optional can\'t be non-greedy AND possessive at the same time'
    s = cur
    if len(input) > 1:
        s += fr'(?:{input})?'
    else:
        if len(input) == 1:
            s += fr'{input}?'
    if not greedy:
        s += r'?'
    if possessive:
        if version_info < (3, 11):
            raise Exception('Possessive qualifiers require at least Python3.11 ')
        s += r'+'
    return s

def _any_of_func(*inputs, chars=None, split=None, cur=...):
    if split and len(inputs) != 1:
        assert False, "Please don't specifiy split and pass multiple inputs to anyof"
    elif split:
        inputs = list(inputs[0])
    elif len(inputs) == 1 and split is None and chars is not False:  # None means auto
        chars = True
        inputs = list(inputs[0])
    elif len(inputs) == 1 and split is None:
        inputs = list(inputs[0])
    elif len(inputs) > 1 and chars is None and all(map(lambda s: len(s) == 1, inputs)):
        chars = True

    if chars:
        cur += r'['
        for i in inputs:
            cur += i
        cur += r']'
    else:
        cur += r'(?:'
        for i in inputs:
            cur += i
            cur += '|'
        cur = cur[:-1]
        cur += r')'
    return cur

def _any_char_except_func(*inputs, cur=...):
    # If it's just a string, split it up
    if len(inputs) == 1 and len(inputs[0]) > 1:
        inputs = list(inputs[0])

    cur += r'[^'
    for i in inputs:
        cur += i
    cur += r']'
    return cur

def _each_func(*inputs, cur=...):
    inputs = list(inputs)
    last = inputs.pop()
    s = cur
    for i in inputs:
        s += fr'(?={i})'
    s += last
    return s


base = {
    # Positional
    'string_starts_with': {'definition': lambda input='', cur=...: r'\A' + input + cur},
    'string_ends_with':   {'definition': lambda input='', cur=...: input + r'\Z' + cur},
    # Always use the multiline flag, so as to distinguish between start of a line vs start of the string
    'line_starts_with':   {'definition': lambda input='', cur=...: r'^' + input + cur, 'flags':'m'},
    'line_ends_with':     {'definition': lambda input='', cur=...: cur + input + r'$', 'flags':'m'},
    'word_boundary':      {'definition': r'\b'},
    'not_word_boundary':  {'definition': r'\B'},

    # Literals
    'tab':                {'definition': r'\t'},
    'space':              {'definition': r' '},
    'space_or_tab':       {'definition': r'[ \t]'},
    'new_line':           {'definition': r'\n'},
    'carriage_return':    {'definition': r'\r'},
    'quote':              {'definition': r'(?:\'|"|`)'},
    'vertical_tab':       {'definition': r'\v'},
    'form_feed':          {'definition': r'\f'},
    'comma':              {'definition': r'\,'},
    'period':             {'definition': r'\.'},
    'underscore':         {'definition': r'_'},
    'any_between':        {'definition': lambda char, and_char, cur=...: cur + r'[' + char + r'-' + and_char + r']'},

    # Not Literals
    'not_whitespace':     {'definition': r'\S'},
    'not_digit':          {'definition': r'\D'},
    'not_word':           {'definition': r'\W'},

    # Catagories
    'whitespace':         {'definition': r'\s'},
    'whitechunk':         {'definition': r'\s+'},
    'digit':              {'definition': r'\d'},
    'number':             {'definition': r'\d+'},
    'word':               {'definition': r'\w+'},
    'word_char':          {'definition': r'\w'},
    'anything':           {'definition': r'.'},
    'chunk':              {'definition': r'.+'},
    'uppercase':          {'definition': r'[A-Z]'},
    'lowercase':          {'definition': r'[a-z]'},
    'letter':             {'definition': r'[A-Za-z]'},
    'hex_digit':          {'definition': r'[0-9a-fA-F]'},
    'oct_digit':          {'definition': r'[0-7]'},
    'punctuation':        {'definition': r'[' + escape('`~!@#$%^&*()-_=+[{]}\\|;:\'",<.>/?¢]') + r']'},
    'controller':         {'definition': r'[\x00-\x1F\x7F]'},
    'printable':          {'definition': r'[\x21-\x7E]'},
    'printable_and_space':{'definition': r'[\x20-\x7E]'},
    'alpha_num':          {'definition': r'[A-Za-z0-9_]'},
    'unicode':            {'definition': lambda name, cur=...: fr'\N{name}'},

    # Amounts
    'match_max':          {'definition': lambda input, cur=...: cur + ('' if not len(input) else r'(?:' + input + r')') + r'+'},
    'match_num':          {'definition': lambda num, input, cur=...: cur + ('' if not len(input) else r'(?:' + input + r')') + r'{' + str(num) + r'}'},
    'match_more_than':    {'definition': lambda min, input, cur=...: cur + ('' if not len(input) else r'(?:' + input + r')') + r'{' + str(int(min) + 1) + r',}'},
    'match_at_least':     {'definition': lambda min, input, cur=...: cur + ('' if not len(input) else r'(?:' + input + r')') + r'{' + str(min) + r',}'},
    'match_at_most':      {'definition': lambda max, input, cur=...: cur + ('' if not len(input) else r'(?:' + input + r')') + r'{0,' + str(max) + r'}'},
    'match_range':        {'definition': _match_range_func},
    'at_least_one':       {'definition': _at_least_one_func},
    'at_least_none':      {'definition': _at_least_none_func},

    # Choices
    'optional':           {'definition': _optional_func},
    'either':             {'definition': lambda input, or_input, cur=...: cur + rf'(?:{input}|{or_input})'},
    'any_of':             {'definition': _any_of_func},
    'any_char_except':    {'definition': _any_char_except_func},
    'any_except':         {'definition': lambda input, type='.*', cur=...: cur + f'(?!{input}){type}'},
    'each':               {'definition': _each_func},

    # Conditionals
    'if_proceded_by':     {'definition': lambda input, cur=...: fr'{cur}(?={input})'},
    'if_not_proceded_by': {'definition': lambda input, cur=...: fr'{cur}(?!{input})'},
    'if_preceded_by':     {'definition': lambda input, cur=...: fr'(?<={input}){cur}'},
    'if_not_preceded_by': {'definition': lambda input, cur=...: fr'(?<!{input}){cur}'},
    'if_enclosed_with':   {'definition': lambda open, stuff, close, cur=...: fr'((?<={open}){stuff}(?={open if close is None else close}))'},

    # Grouping
    'group':              {'definition': lambda input, name=None, cur=...: f'{cur}({input})' if name is None else f'{cur}(?P<{name}>{input})'},
    'passive_group':      {'definition': lambda input, cur=...: f'{cur}(?:{input})'},
    'earlier_group':      {'definition': lambda num_or_name, cur=...: f'{cur}\\{num_or_name}' if isinstance(num_or_name, int) or num_or_name in digits else f'{cur}(?P={num_or_name})'},
    'if_exists':          {'definition': lambda num_or_name, does, doesnt, cur=...: f'{cur}(?({num_or_name}){does}{"|" + str(doesnt) if doesnt is not None else ""})'},

    # Premade
    'literally_anything':  {'definition': r'(?:.|\n)'},
    'signed':             {'definition': r'(?:(?:\-|\+))?\d+'},
    'unsigned':           {'definition': r'\d+'},
    'plain_float':        {'definition': r'(?:(?:\-|\+))?\d+\.(?:\d+)?'},
    'full_float':         {'definition': r'(?:(?:\-|\+))?\d+\.(?:\d+)?(?:e(?:(?:\-|\+))?\d+)?'},
    'int_or_float':       {'definition': r'(?:(?:\-|\+))?\d+\.(?:\d+)?(?:e(?:(?:\-|\+))?\d+)?(?:\-)?\d+(?:\.(?:\d+)?)?'},
    'ow':                 {'definition': r'(?:\s+)?'},

    # Misc.
    'is_exactly':         {'definition': lambda input, cur=...: r"\A" + input + r'\Z'},
    'literal':            {'definition': lambda input, cur=...: cur + input},
    'raw':                {'definition': lambda regex, cur=...: str(regex), 'sanatize': False},

    # Flags
    'ASCII':              {'definition': lambda cur=...: cur, 'flags': 'a'},
    'DOTALL':             {'definition': lambda cur=...: cur, 'flags': 's'},
    'IGNORECASE':         {'definition': lambda cur=...: cur, 'flags': 'i'},
    'LOCALE':             {'definition': lambda cur=...: cur, 'flags': 'L'},
    'MULTILINE':          {'definition': lambda cur=...: cur, 'flags': 'm'},
    'UNICODE':            {'definition': lambda cur=...: cur, 'flags': 'u'},
}



from inspect import currentframe, getframeinfo

from _regexs import *

import ezregex.python as er
from ezregex.invert import *
from ezregex.python import *

# This goes (regex,                                                                         (things it should match),                                    (things it shouldnt match))
_regexsLine = getframeinfo(currentframe()).lineno + 2
regexs = (
    ('stuff' + ' ' + optional(comma) + space + ifFollowedBy('*'),                           ['stuff , *', 'stuff  *'],                                              None),
    ('stuff' + anyof('a', 'b', 'c') + ' ' + optional(comma) + space + ifFollowedBy('*'),    ['stuffa , *', 'stuffb  *'],                                            None),
    (anyof('a', 'b', 'c') + ' ' + optional(comma) + space + ifFollowedBy('*'),              ['a , *', 'b  *'],                                                      None),
    ('stuff' + anyof('a', 'b', 'c'),                                                        ['stuffa', 'stuffb'],                                                   None),
    (anyof('a', 'b', 'c'),                                                                  ['a', 'b', 'c'],                                                        None),
    (one_of('a', 'b', 'c'),                                                                 ['a', 'b', 'c'],                                                        None),
    ('a' + ifFollowedBy('*'),                                                               ['a*'],                                                                 None),
    (optional(comma) + space,                                                               [', ', ' '],                                                            None),
    (optional(word + ow + ',' + ow) + group(word) + optional(',') + ow,                     ['word\t ,word2, ', 'word', 'worddsfs    ', 'word,   '],                ('', '  ')),
    (optional(whitechunk + 'as' + word),                                                    [' asword'],                                                            None),
    (group(optional(matchMax(er.literal('.') + word))),                                     ['..........word', ''],                                                 None),
    (matchNum(3, either("'", '"')),                                                         ['"""'],                                                                None),
    (matchNum(2, '0'),                                                                      ['00'],                                                                 None),
    (matchNum(1, raw('6')),                                                                 ['6'],                                                                  None),
    (raw(r'\A'),                                                                            [''],                                                                   None),
    (raw(r'\Z'),                                                                            [''],                                                                   None),
    (exactly('test'),                                                                       ['test'],                                                               None),
    (raw(r'(test)+'),                                                                       ['test', 'testt', 'testtt', 'testttt'],                                 None),
    (raw(r'test+'),                                                                         ['testtesttest'],                                                       None),
    (raw(r'(test){3}'),                                                                     ['testtesttest'],                                                       None),
    (raw(r'test{3}'),                                                                       ['testttt'],                                                            None),
    (raw(r'(test){3,5}'),                                                                   ['testtesttest', 'testtesttesttesttest'],                               None),
    (raw(r'(test){3, 5}'),                                                                  None,                                                                   None),
    (raw(r'test{3, 5}'),                                                                    None,                                                                   None),
    (raw(r'test{3,5}'),                                                                     ['testttt', 'testttttt'],                                               None),
    (raw(r'(test){3,}'),                                                                    ['testtesttest', 'testtesttesttest', 'testtesttesttesttest'],           None),
    (raw(r'test{3,}'),                                                                      ['testttt', 'testtttt', 'testttttt'],                                   None),
    (raw(r'\d'),                                                                            ['0', '1', '2'],                                                        None),
    (raw(r'\s'),                                                                            [' ', '\t', '\n'],                                                      None),
    (raw(r'\w'),                                                                            ['a', 'Z', '5', '_'],                                                   None),
    (raw(r'\w+'),                                                                           ['abc', '123', 'test_'],                                                None),
    (raw(r'.'),                                                                             'sahrdhcfGSGD87w3ue84125asd_;.,?.134*&^`'.split(),                      None),
    (raw(r'\n'),                                                                            ['\n'],                                                                 None),
    (raw(r'\r'),                                                                            ['\r'],                                                                 None),
    (raw(r'\t'),                                                                            ['\t'],                                                                 None),
    (raw(r'\v'),                                                                            ['\v'],                                                                 None),
    (raw(r'\f'),                                                                            ['\f'],                                                                 None),
    (raw(r'\S'),                                                                            'sdgSGHR5122$%^&*Z`'.split(),                                           None),
    (raw(r'\D'),                                                                            'sdfGSDG;([]'.split(),                                                  None),
    (raw(r'\W'),                                                                            '/*-#^&*`?><'.split(),                                                  None),
    (raw(r'(stuff)?'),                                                                      ['stuff', ''],                                                          None),
    (raw(r'(stuff)*'),                                                                      ['', 'stuff', 'stuffstuff', 'stuff'*3],                                 None),
    (raw(r'(stuff|things)'),                                                                ['stuff', 'things'],                                                    None),
    (raw(r'[s,t]'),                                                                         ['s', 't'],                                                             None),
    (raw(r'[^s,t]'),                                                                        'qwruiopahjkBNM34#$'.split(),                                           None),
    (raw(r'(?=stuff)'),                                                                     ['stuff'],                                                              None),
    (raw(r'(?!stuff)'),                                                                     ['stuff'],                                                              None),
    (raw(r'(stuff)'),                                                                       ['stuff'],                                                              None),
    (raw(r'(stuff)?'),                                                                      ["stuff", ''],                                                          None),
    (raw(r'(?P<name>stuff)'),                                                               ['stuff'],                                                              None),
    (raw(r'(?<=stuff)'),                                                                    ['thingstuffs'],                                                        None),
    (raw(r'(?<!stuff)'),                                                                    ['thingstuffs'],                                                        None),
    (raw(r'(a|b|c|thing|st)uff'),                                                           ["auff", "buff", "cuff", "thinguff", "stuff"],                          None),
    (raw(r"A AB ABC [^A] [^ABC] A+ A* A? AA* A{2} A{2,4} A{9,12} A{2,} A{,9}"),             None,                                                                   None), # I found these online for testing a different regex suite.
    (raw(r"[\d] [^\d] [A-Za-z0-9_]+ . .*"),                                                 None,                                                                   None), # Figured they'd make good tests.
    (raw(r"AB|CD AB|CD|EF|GH (AB|CD)*"),                                                    None,                                                                   None),
    (raw(r"(a??) a*? a{3,}? ab{4,7}?"),                                                     None,                                                                   None),
    (raw(r"(?P<test>ABC*) (?P<a>x)|(?P<b>y)"),                                              None,                                                                   None),
    (raw(r"[b:]+ (b)|(:+) a|(b)"),                                                          None,                                                                   None),
    (raw(r"(?:(?P<a1>a)|(?P<b2>b))(?P<c3>c)?"),                                             None,                                                                   None),
    (raw(r"(?P<name>[a-zA-Z]+)(?P=name)"),                                                  None,                                                                   None),
    (group(+letter, name='g') + earlierGroup('g'),                                          ('AA', 'tt'),                                                           ('ABt','t', '9d9', 'tdt')),
    (group(+letter) + ' ' + earlierGroup(1),                                                ('the the', 'at at'),                                                   ('att', 'thethe')),
    (raw(r"[AB\]C] [--A] [ABC\-D] [\^ABC]"),                                                None,                                                                   None),
    (stringStartsWith('a'),                                                                 ('asdfs', 'a 89sdf a', 'a'),                                            (' asdf', 'sdfa', 'sdf')),
    (lineStartsWith('a'),                                                                   ('asdfs', 'a 89sdf a', 'a', 'sdfs\nasdfd'),                             (' asdf', 'sdf\nsdf', 'sdf\n a', 'sdfa', 'sdf')),
    (stringEndsWith('a'),                                                                   ('lklkjfda', 'sdf 8 a', 'a'),                                           ('asdfds', 'sd fd')),
    (lineEndsWith('a'),                                                                     ('lklkjfda', 'sdf 8 a', 'a', 'sdf\na', 'sdfse\nsdafsda', 'sdfa\nsdf'),  ('asdfds', 'sd fd')),
    (er.literal('test'),                                                                    ('test', ' sdfstestsdfs',),                                             ('te st',)),
    (underscore,                                                                            ('_',),                                                                 ('a test', 'test ', '\ntest\n', '\ntest', 'test\n')),
    (quote,                                                                                 ('"', "'", '`'),                                                        ('a test', 'test ', '\ntest\n', '\ntest', 'test\n')),
    (isExactly('test'),                                                                     ('test',),                                                              ('a test', 'test ', '\ntest\n', '\ntest', 'test\n')),
    (matchMax('a'),                                                                         ('aaa', 'a'),                                                           ('b',)),
    (matchMoreThan(3, 'a'),                                                                 ('aaaa', 'tesaaaaaat'),                                                 ('aaa',' aa')),
    (matchAtLeast(3, 'a'),                                                                  ('aaa', 'aaaa'),                                                        ('aa',)),
    (optional('a'),                                                                         ('', 'a', 'aa'),                                                        None),
    (either('a', 'b'),                                                                      ('a', 'b'),                                                             ('c',)),
    (either('aa', 'ba'),                                                                    ('aa', 'ba',),                                                          ('bb', 'a')),
    (whitespace,                                                                            (' ', '\t', '\t  ', '\n'),                                              ('dfsd',)),
    (whitechunk,                                                                            (' ', '\t', '\t  ', '\n'),                                              ('dfsd',)),
    (white,                                                                                 (' ', '\t', '\t  ', '\n'),                                              ('dfsd',)),
    (digit,                                                                                 ('6',),                                                                 ('_', '-', 'a')),
    (number,                                                                                ('6', '69'),                                                            ('-a', 'A')),
    (punctuation,                                                                           '@#$%^&*()'.split(),                                                    '12345678sdfsdf'.split()),
    (wordChar,                                                                              ('w',),                                                                 ('-',)),
    (hex_digit,                                                                              ('A', 'a', '0'),                                                        ('g', 'G')),
    (octDigit,                                                                              ('7',),                                                                 ('9', 'a', 'A', '8')),
    (chunk,                                                                                 ('wordssdf   asdf\n',),                                                 ('\n',)),
    (spaceOrTab,                                                                            (' ', '\t', ' \t  '),                                                   ('\n',)),
    (newLine,                                                                               ('\n',),                                                                ('\r',)),
    (carriageReturn,                                                                        ('\r',),                                                                ('\n',)),
    (tab,                                                                                   ('\t',),                                                                (' ',)),
    (space,                                                                                 (' ',),                                                                 ('\t',)),
    (quote,                                                                                 ('"', "'"),                                                             ('wer',)),
    (comma,                                                                                 (',',),                                                                 ('.', '`')),
    (period,                                                                                ('.',),                                                                 (',',)),
    (matchRange(3, 5, 'a'),                                                                 ('aaa', 'aaaa', 'a'*5, 'a'*6),                                          ('aa',)),
    (matchRange(3, 5, 'a', greedy=False)+'aa',                                              ('a'*5, 'a'*6),                                                         ('aa',)),
    (optional('a') + 'b',                                                                   ('b', 'ab','cb'),                                                       ('','c')),
    (optional('a', greedy=False) + 'b',                                                     ('b', 'ab','cb'),                                                       ('','c')),
    (atLeastOne('a'),                                                                       ('a','aa', 'a'*20),                                                     ('', 'b')),
    (atLeastOne('a', greedy=False),                                                         ('a','aa', 'a'*20),                                                     ('', 'b')),
    (atLeastNone('a'),                                                                      ('', 'a', 'a'*20, 'b'),                                                 None),
    ((optional('a') + 'b') * 3,                                                             ('abbb', 'bbb', 'ababab', 'bbab'),                                      ('', 'aaa', 'aa', 'a')),
    (word + whitechunk + group('func') + ':' + '()' + group(either('|', '7'), name='test'), ('wo  func:()|', 'wo  func:()7'),                                       None),
    (word + whitechunk + group('func') + ':' + group(anyof('8', '7'), 'test'),              ('wo  func:8', 'wo  func:7'),                                           None),
    (7 + anyof('abc') + lineEnd,                                                            ('7a', 'sdfsd7b', 'sdf\nsdf7b', 'sdf\nsdf7b\n'),                        ('7asdfsd', '7v')),
    (7 + anyof('abc') + stringEnd,                                                          ('7a', 'sdfsd7b'),                                                      ('7asdfsd', '7v', 'sdf\nsdf7bds', 'sdf\nsdf7bf\n')),
    (lineStart + 7 + anyof('abc'),                                                          ('7a', '7bsdfsd', '\n7a', '\n7bsdfsd'),                                 ('ds7asdfsd', '7v')),
    (stringStart + 7 + anyof('abc'),                                                        ('7a', '7bsdfsd'),                                                      ('ds7asdfsd', '7v', '\n7a', '\n7bsdfsd')),
    (+alpha,                                                                                ('a', 'asd'),                                                           ('89', '._78')),
    (+alphanum,                                                                             ('a', 'asd', '3sd', '88'),                                              ('.+',)),
    (raw(r'(<)?(\w+@\w+(?:\.\w+)+)(?(1)>|$)'),                                              ('<user@host.com>', 'user@host.com'),                                   ('user@host.com>',)), # https://docs.python.org/3/library/re.html says '<user@host.com' shouldn't match this pattern, but it's wrong...
    ('foo' + each(chunk + 'here' + chunk, chunk + anyOf('this', 'that') + chunk) + 'bar',   ('fooum here there that bar', 'foo that there here bar'),               None),
    # ('foo' + ((chunk + 'here' + chunk) & (chunk + anyOf('this', 'that') + chunk)) + 'bar',  ('fooum here there that bar', 'foo that there here bar'),               None),
    # ('foo' + ((chunk + 'here' + chunk) & (chunk + anyOf('this', 'that') + chunk) & (chunk + 'the' + chunk)) + 'bar',  ('fooum the here there that bar', 'foo that there thehere bar'), ('fooum here there that bar', 'foo that there here bar')),
    (opt(group('<')) + group(word + '@' + word + +('.' + word)) + ifExists(1, '>', string_end), ('<user@host.com>', 'user@host.com'),                               ('user@host.com>',)), # https://docs.python.org/3/library/re.html says '<user@host.com' shouldn't match this pattern, but it's wrong...
    # (,                                                                                    (,),                                                                    (,)),
    # (word_boundary + word_char[...,3] + wordBoundary,                                       ('yes', 'hey', 'sup', 'thi'),                                           ('none', 'no', 'foo3', '333', 'jar_')), # This does what it's supposed to do, but doesn't search correctly, as far as I understand it
    #TODO ('foo ' + anyExcept('bar') + ' baz',                                               ('foo thing baz', 'foo bax baz'),                                       ('foo bar baz',)),
    #TODO: (exactly('foo' + anyExcept('boo') + 'bar'),                                             ('foonotbar', 'foobar'),                                                ('fooboobar', 'boobar', 'fooboo')), # not sure where 'foo boo bar' goes
    #TODO: (exactly('foo' + anyExcept('boo', number) + 'bar'),                                     ('foo999bar', 'foo8bar'),                                               ('fooboobar','foonotbar', 'foo boo bar', 'foobar', 'boobar')),
    #TODO: (exactly('foo' + anyExcept('boo', match_amt(3, number)) + 'bar'),                       ('foo999bar', 'foo989bar'),                                             ('fooboobar','foonotbar', 'foo boo bar', 'foobar', 'boobar', 'foo8bar', 'foo98')),
    #TODO: (exactly(matchAtMost(3, digit)),                                                        ('444', '33', '1'),                                                     ('9999','9830')),
    #TODO: (exactly('foo' + anyExcept('888', match_amt(3, number)) + 'bar'),                       ('foo999bar', 'foo989bar'),                                             ('fooboobar','foonotbar', 'foo boo bar', 'foobar', 'foo888bar', 'boobar', 'foo8bar', 'foo98')),
    #TODO: (exactly('foo' + anyExcept('boo', word) + 'bar'),                                       ('foonotbar',),                                                         ('fooboobar', 'foo99bar', 'fooboo', 'boobar', 'foobar')),
    # TODO: punctuation
    # TODO: anyExcept(punctuation)
    # TODO: anyExcept(anyof(punctuation))
    # (,                                                                                    (,),                                                                    (,)),
    # TODO:
    # (raw(r"[ABC]+(?=D).*$ <.*?>"),                                                          None,                                                                   None),
    # (raw(r"(?:Q)(Q) ^A*$"),                                                                  None,                                                                   None), # I'm not actually sure what this is supposed to do
    # (raw(r"(?=AB)C (?!CD)DC AB(?<=CD) AB(?<!CD)"),                                           None,                                                                   None), # also not sure how this is actually supposed to work
    # (matchRange(3, 5, 'a', possessive=True) + 'aa',                                       ('a'*7,),                                                               ('a'*6,)),
    # (optional('a', possessive=True) + 'b',                                                ('',),                                                                  ('',)),
    # (atLeastOne(possessive=True),                                                         ('',),                                                                  ('',)),
    # (atLeastNone(possessive=True),                                                        ('',),                                                                  ('',)),
    # verticalTab,                                                                          ('',),                                                                  ('',),
    # formFeed,                                                                             ('',),                                                                  ('',),
    # either('a', 'b'),                                                                     ('a', 'b'),                                                             None),
    # multiOptional,                                                                        ('',),                                                                  ('',),
    # anyBetween,                                                                           ('',),                                                                  ('',),
    # (unicode,                                                                             ('',),                                                                  ('',),
    # (word,                                                                                ('word',),                                                              ('33a',)), # Is this *supposed* to work?
)

# This goes ("regex pattern", "replacement regex", "base string", "what the base string should look like after substitution"),
replacements = (
    (digit,                         "*",        "123abc",   "***abc"),  # Replace digits with *
    (anyof('aeiou'),                "_",        "hello",    "h_ll_"),  # Replace vowels with _
    (digit,                         "#",        "a1b2c3",   "a#b#c#"),  # Replace digits with #
    (group(digit) + group(digit),   rgroup(2) + rgroup(1),  "12345", "21435"),  # Swap the first two digits
    (group(anyof('aeiou')),         rgroup(1) * 2,          "hello", "heelloo"),  # Duplicate vowels
    (group(number),                 "num:" + rgroup(1),     "abc123xyz456", "abcnum:123xyznum:456"),  # Prefix numbers with "num:"
    (group(word_char) * 2,          rgroup(1) + '-' + rgroup(2), "Python", "P-yt-ho-n"),  # Separate adjacent characters with -
    (group(+uppercase),             rgroup(1) * 2,          "HelloWorld", "HHelloWWorld"),  # Duplicate uppercase words
    (group(+not_word),              " ",        "a@b#c",    "a b c"),  # Replace non-word characters with space
    # ("(\\b\\w{3}\\b)", "[\\1]", "The cat sat on the mat", "The [cat] [sat] on the [mat]"),  # Enclose 3-letter words in square brackets
    (group(digit) + group(lowercase), rgroup(1) + '_' + rgroup(2), "1a2b3c", "1_a2_b3_c"),  # Add underscore between digits and lowercase letters
    (group(+not_whitespace),        f"<{rgroup(1)}>", "Hello World", "<Hello> <World>"),  # Enclose non-space sequences in angle brackets
    # ("(\\b\\w+\\b)", "\\U\\1", "python is fun", "PYTHON IS FUN"),  # Convert words to uppercase
    (group(digit, 'digit') + group(lowercase, 'letter'), rgroup('letter') + rgroup('digit'), "1a2b3c", "a1b2c3"),  # Swap digit and letter with named groups
    (group(anyof('aeiou'), 'vowel'), f"[{rgroup('vowel')}]", "hello", "h[e]ll[o]"),  # Enclose vowels in square brackets with named group
    # ("(?P<number>[0-9]+)", "num: \\g<number>", "abc123xyz456", "abcnum: 123xyznum: 456"),  # Prefix numbers with "num:" using named group
    # ("(?P<upper>[A-Z]+)", "\\g<upper>\\g<upper>", "HelloWorld", "HELLOHELLO"),  # Duplicate uppercase words with named group
    # ("(?P<word>\\w+)", "<\\g<word>>", "Python is great", "<Python> <is> <great>"),  # Enclose words in angle brackets with named group
    ('foo' + group(number, 'num') + 'bar', f"foo-{rgroup('num')}-bar", 'foo87bar', 'foo-87-bar'),
    # ("(?P<non_space>\\S+)", "<\\g<non_space>>", "Hello World", "<Hello> <World>"),  # Enclose non-space sequences in angle brackets with named group
    # ("(?P<word>\\b\\w+\\b)", "\\U\\g<word>", "python is cool", "PYTHON IS COOL"),  # Convert words to uppercase with named group
    (group(word + number) + ':' + ow + group(word), replace_group(1) + ' - ' + replace_group(2), 'test1:    thing', 'test1 - thing'),
    (group(word + number, 'a') + ':' + ow + group(word, 'b'), replace_group('a') + ' - ' + replace_group('b'), 'test1:    thing', 'test1 - thing'),
    (stringStart + '(' + group(chunk + optional(',' + chunk)) + ')' + chunk, '(' + '${' + rgroup(1) + '})', '(name, input) -> ezregex.EZRegex.EZRegex', '(${name, input})'),
)



import json
import string
import traceback
from random import choice, choices, randint
from re import search
from sys import version_info
from typing import Literal, Union

from ezregex import *

if version_info.minor <= 10:
    from re import sre_parse as sre  # type: ignore
else:
    from re import _parser as sre  # type: ignore

# So I can debug this function directly
if __name__ != '__main__':
    from .invert_old import invertRegex

with open('ezregex/assets/words.json') as f:
    words = json.load(f).keys()

# Don't use string.whitespace, because we don't want to use weird difficult to print characters.
# We want the replacement to be readable.
_whitespace   = ' '
_everything   = string.digits + string.ascii_letters + string.punctuation + _whitespace + '_'

def _randWord(length=None, word='lookup'):
    if word == 'random':
        return ''.join(choices(string.ascii_letters + '_', k=length or randint(3, 9)))
    elif word == 'lookup':
        options = list(filter((lambda i: len(i) > 2 and len(i) < 10) if length is None else (lambda i: len(i) == length), words))
        if not len(options):
            _randWord(length, word='random')
        return choice(options)
    elif word is None:
        return 'word'
    else:
        raise ValueError(f"invalid parameter given for word {word}. Accepted values are either random, lookup, or None")

def _randNumber(length:int, random=False) -> str:
    if random:
        return str(randint(0, 10**length))
    else:
        return ''.join(map(lambda i: str(i)[-1], range(1, length+1)))

def invert(
    expr:Union[str, 'EZRegex'],
    words:Literal['lookup', 'random']='lookup',
    randomNumbers=False,
    alot=8,
    tries:int=10,
    backend:Literal['re_parser', 'regex', 'xeger', 'sre_yield']='re_parser',
    _verbose=False,
    # Here for recursion; not to be used
    _untried=['re_parser', 'regex', 'xeger', 'sre_yield'],
    _tried=...,
) -> str:
    """ "Inverts" a regular expression by returning an example of something which is guaruanteed to
        match the passed expression.
        NOTE: This only works on valid Python regular expressions (it will probably mostly work for
            other dialects, but is not guarenteed.)
        expr: The regular expression to invert. Can be a string, or a EZRegex expression
        words: Controls how works are handled. If `random`, words are made of random letters. If `lookup`,
            it looks up valid english words and inserts them to make it more readable.
        randomNumbers: controls whether all numbers are 12345... to a desired length, or if they're
            just random numbers (again, for readability)
        alot: When given a choice of how many characters to put someone, it inserts a random integer
            between 1 and `alot`.
        tries: Controls how many times to try to invert the expression before giving up. This is effective,
            because there is an element of randomness involved in inverting the regex given.
        backend: One of ('re_parser', 'regex', 'xeger', 'sre_yield').
            `re_parser` is the default, it uses the build-in parser in the re package to create an AST
                of the regex
            `regex` uses regular expressions to parse regular expressions, which is as gross as it sounds.
                It's slightly buggy, but mostly works.
            `xeger` imports the `xeger` pacakge and uses it instead. Xeger inverts work, but are less
                readable. Must have the `xeger` package installed.
            `sre_yield` imports the `sre_yield` package and uses it instead. I don't think this works
                right now, and may be significantly slower.
    """
    expr = str(expr)
    if _tried is Ellipsis:
        _tried = tries

    # If we get an error in the actual inversion part, just handle it as though
    # we generated a bad string that doesn't match, and continue recusing
    try:
        match backend:
            case 'xeger':
                from xeger import Xeger  # type: ignore
                rtn = Xeger().xeger(expr)
            case 'regex':
                # Tries already gets implemented in this function
                try:
                    rtn = invertRegex(expr, 1)
                except NotImplementedError:
                    # Handle the error ourselves
                    rtn = None
            case 'sre_yield':
                import sre_yield  # type: ignore
                for i in sre_yield.AllStrings(expr):
                    rtn = i
            case 're_parser':
                groups = {}
                def handle(pattern, amt=1, opposite=False):
                    if _verbose:
                        print(f'Handling {pattern} * {amt}')
                    s = ''
                    for op, args in pattern:
                        if not opposite:
                            match op:
                                case sre.LITERAL:
                                    s += chr(args) * amt
                                case sre.NOT_LITERAL:
                                    almost_everything = list(_everything)
                                    if chr(args) in almost_everything:
                                        almost_everything.remove(chr(args))
                                    # If it's not in there, great!
                                    s += choice(almost_everything) * amt
                                case sre.RANGE:
                                    start, end = args
                                    s += chr(randint(start, end))
                                case sre.MAX_REPEAT | sre.MIN_REPEAT:
                                    min, max, sub = args
                                    if max is None or max is sre.MAXREPEAT:
                                        max = randint(min, alot)
                                    try:
                                        # If we know we're getting word chars, make a word of it
                                        if type(sub[0][1][0]) is not int and sub[0][1][0][1] in (sre.CATEGORY_WORD, sre.CATEGORY_UNI_WORD, sre.CATEGORY_LOC_WORD):
                                            if _verbose: print('Getting a random word')
                                            s += _randWord(word=words)
                                        elif type(sub[0][1][0]) is not int and sub[0][1][0][1] in (sre.CATEGORY_DIGIT, sre.CATEGORY_UNI_DIGIT):
                                            if _verbose: print('Getting whole number')
                                            s += _randNumber(randint(1, alot), random=randomNumbers)
                                        else:
                                            s += handle(sub, randint(min, max) * amt)
                                    except Exception as err:
                                        if _verbose:
                                            print(err)
                                            print(traceback.format_exc())

                                        s += handle(sub, randint(min, max) * amt)
                                case sre.ANY:
                                    for _ in range(amt):
                                        s += choice(_everything)
                                case sre.ASSERT:
                                    type_, sub = args
                                    if type_ == 1: # ifProcededBy
                                        s = s + handle(sub, amt)
                                    elif type_ == -1: # ifPrecededBy
                                        s += handle(sub, amt)
                                    else:
                                        s += handle(sub, amt)
                                case sre.ASSERT_NOT:
                                    type_, sub = args
                                    if type_ == 1: # ifNotProcededBy
                                        # almost_everything = list(_everything)
                                        # if chr(args) in almost_everything:
                                        #     almost_everything.remove(chr(args))
                                        # # If it's not in there, great!
                                        # s += choice(almost_everything) * amt
                                        s += handle(sub, amt, opposite=True)
                                    if type_ == -1: # ifNotPrecededBy
                                        s = s + handle(sub, amt, opposite=True)
                                    # If it needs to be followed by that sequence, simply add that sequence
                                    else:
                                        s += handle(sub, amt, opposite=True)
                                        # s += handle(sub, amt)
                                case sre.CATEGORY:
                                    for _ in range(amt):
                                        match args:
                                            case sre.CATEGORY_DIGIT | sre.CATEGORY_UNI_DIGIT:
                                                s += str(randint(0, 10))
                                            case sre.CATEGORY_LINEBREAK | sre.CATEGORY_UNI_LINEBREAK:
                                                s += '\n'
                                            case sre.CATEGORY_NOT_DIGIT | sre.CATEGORY_UNI_NOT_DIGIT:
                                                s += choice(string.ascii_letters + string.punctuation + _whitespace)
                                            case sre.CATEGORY_NOT_LINEBREAK | sre.CATEGORY_UNI_NOT_LINEBREAK:
                                                s += choice(_everything)
                                            case sre.CATEGORY_NOT_SPACE | sre.CATEGORY_UNI_NOT_SPACE:
                                                s += choice(string.punctuation + string.ascii_letters + string.digits)
                                            case sre.CATEGORY_NOT_WORD | sre.CATEGORY_UNI_NOT_WORD | sre.CATEGORY_LOC_NOT_WORD:
                                                s += choice(string.punctuation + _whitespace)
                                            case sre.CATEGORY_SPACE | sre.CATEGORY_UNI_SPACE:
                                                s += ' '
                                            case sre.CATEGORY_WORD | sre.CATEGORY_UNI_WORD | sre.CATEGORY_LOC_WORD:
                                                s += choice(string.ascii_letters + '_')
                                            case _:
                                                raise NotImplementedError(f'Unknown category given: {args}')
                                case sre.IN:
                                    # If this is a pattern like [^abc]
                                    if args[0][0] is sre.NEGATE:
                                        # Handle all the args except the negate flag and remove it from the options
                                        otherthan = handle(args[1:])
                                        almost_everything = list(_everything)
                                        for i in otherthan:
                                            if i in almost_everything:
                                                almost_everything.remove(i)
                                        s += choice(almost_everything)
                                    else:
                                        s += handle([choice(args)], amt)
                                case sre.SUBPATTERN: # This handles groups
                                    group, num, num2, sub = args
                                    sub_pattern = handle(sub) * amt
                                    groups[group] = sub_pattern
                                    s += sub_pattern
                                case sre.AT:
                                    if args is sre.AT_BEGINNING_STRING:
                                        # Whatever comes next better be first
                                        s = ''
                                    elif args is sre.AT_END_STRING:
                                        # Whatever comes next better be last
                                        return s
                                    elif args is sre.AT_BEGINNING:
                                        s += '\n'
                                    elif args is sre.AT_END:
                                        s += '\n'
                                    elif args is sre.AT_BOUNDARY:
                                        if len(s) and s[-1] in string.digits + string.ascii_letters + '_':
                                            s += choice(string.punctuation + _whitespace)
                                        else:
                                            s += choice(string.digits + string.ascii_letters + '_')
                                    else:
                                        raise NotImplementedError(f'Unknown parameter[s] given for AT op: {args}')
                                case sre.BRANCH:
                                    something, branches = args
                                    s += handle(choice(branches), amt)
                                case sre.NEGATE:
                                    s += handle(args, amt, opposite=True)
                                case sre.GROUPREF:
                                    s += groups[args]
                                case sre.GROUPREF_EXISTS:
                                    # TODO: This often requires multiple tries to get right
                                    group, trueSub, falseSub = args
                                    s += handle(trueSub if group in groups else falseSub, amt)
                                case _:
                                    raise NotImplementedError(f'Unknown op {op} given with args {args}')
                        else:
                            match op:
                                case sre.LITERAL:
                                    almost_everything = list(_everything)
                                    if chr(args) in almost_everything:
                                        almost_everything.remove(chr(args))
                                    # If it's not in there, great!
                                    s += choice(almost_everything) * amt
                                case sre.NOT_LITERAL:
                                    s += chr(args) * amt
                                case sre.RANGE:
                                    start, end = args
                                    almost_everything = set(_everything)
                                    almost_everything.difference(map(chr, range(start, end)))
                                    s += choice(list(almost_everything))
                                case sre.MAX_REPEAT | sre.MIN_REPEAT:
                                    min, max, sub = args
                                    # Adding something that wouldn't match at all is always an option
                                    options = [handle(sub, amt, opposite=True)]
                                    # If a max is specified, then it can't match more than that many
                                    if max is not None and max is not sre.MAXREPEAT:
                                        options.append(handle(sub, amt * (max + randint(1, alot))))
                                    # If there's a min (meaning its more than 0), then it can't match
                                    # less than that many
                                    if min >= 1:
                                        # TODO: Potential Error: This doesn't take into account the current amt.
                                        options.append(handle(sub, randint(0, min-1)))
                                    s += choice(options)
                                case sre.ANY:
                                    # TODO: I think this will fail if given [^.\\n] (or anyExcept(literallyAnything)),
                                    # but also, how DO you handle that?
                                    for _ in range(amt):
                                        # I guess?
                                        s += '\n'
                                case sre.ASSERT:
                                    # I don't think conditionals are allowed inside a not assert (I think)
                                    type_, sub = args
                                    s += handle(sub, amt, opposite=True)
                                case sre.ASSERT_NOT:
                                    type_, sub = args
                                    # I *think* this will work?...
                                    s += handle(sub)
                                case sre.CATEGORY:
                                    for _ in range(amt):
                                        match args:
                                            case sre.CATEGORY_DIGIT | sre.CATEGORY_UNI_DIGIT:
                                                s += choice(string.ascii_letters + string.punctuation + _whitespace)
                                            case sre.CATEGORY_LINEBREAK | sre.CATEGORY_UNI_LINEBREAK:
                                                s += choice(_everything)
                                            case sre.CATEGORY_NOT_DIGIT | sre.CATEGORY_UNI_NOT_DIGIT:
                                                s += choice(string.digits)
                                            case sre.CATEGORY_NOT_LINEBREAK | sre.CATEGORY_UNI_NOT_LINEBREAK:
                                                s += '\n'
                                            case sre.CATEGORY_NOT_SPACE | sre.CATEGORY_UNI_NOT_SPACE:
                                                s += ' '
                                            case sre.CATEGORY_NOT_WORD | sre.CATEGORY_UNI_NOT_WORD | sre.CATEGORY_LOC_NOT_WORD:
                                                s += choice(string.ascii_letters + '_')
                                            case sre.CATEGORY_SPACE | sre.CATEGORY_UNI_SPACE:
                                                s += choice(string.punctuation + string.ascii_letters + string.digits)
                                            case sre.CATEGORY_WORD | sre.CATEGORY_UNI_WORD | sre.CATEGORY_LOC_WORD:
                                                s += choice(string.punctuation + _whitespace)
                                            case _:
                                                raise NotImplementedError(f'Unknown category given: {args}')
                                case sre.IN:
                                    # If this is a pattern like [^abc]
                                    if args[0][0] is not sre.NEGATE:
                                        # Handle all the args except the negate flag and remove it from the options
                                        otherthan = handle(args[1:])
                                        almost_everything = list(_everything)
                                        for i in otherthan:
                                            if i in almost_everything:
                                                almost_everything.remove(i)
                                        s += choice(almost_everything)
                                    else:
                                        s += handle([choice(args)], amt)
                                case sre.SUBPATTERN:
                                    idk, num, num2, sub = args
                                    s += handle(sub, opposite=True) * randint(0, amt)
                                case None: #sre.AT: # I don't know how to handle this
                                    if args is sre.AT_BEGINNING_STRING:
                                        # Whatever comes next better be first
                                        s = ''
                                    elif args is sre.AT_END_STRING:
                                        # Whatever comes next better be last
                                        return s
                                    elif args is sre.AT_BEGINNING:
                                        s += '\n'
                                    elif args is sre.AT_END:
                                        s += '\n'
                                    elif args is sre.AT_BOUNDARY:
                                        if len(s) and s[-1] in string.digits + string.ascii_letters + '_':
                                            s += choice(string.punctuation + _whitespace)
                                        else:
                                            s += choice(string.digits + string.ascii_letters + '_')
                                    else:
                                        raise NotImplementedError(f'Unknown parameter[s] given for AT op: {args}')
                                case sre.BRANCH:
                                    something, branches = args
                                    s += handle(choice(branches), amt, opposite=True)
                                case sre.NEGATE:
                                    s += handle(args, amt)
                                case sre.GROUPREF:
                                    add = groups[args]
                                    while add == groups[args]:
                                        add = _randWord(word=words)
                                    s += add
                                case sre.GROUPREF_EXISTS:
                                    group, trueSub, falseSub = args
                                    s += handle(trueSub if group in groups else falseSub, amt, opposite=True)
                                case _:
                                    raise NotImplementedError(f'Unknown opposite op {op} given with args {args}')
                    return s

                rtn = handle(sre.parse(expr))
    except:
        rtn = None
    # I don't know why copying here is necissary. Do parameters act like globals?
    _untried = _untried.copy()
    try:
        _untried.remove(backend)
    # If we get a ValueError here, it just means that we're retrying the current backend
    # We can't just remove the _untried parameter in the recuse call in the else
    # statement below, because then it causes an infinite loop, instead of trying all
    # the options. This is better.
    except ValueError: pass

    # Try the current backend `tries` times
    if rtn is not None and search(expr, rtn):
        return rtn
    # If we're out of tries, switch backends
    elif tries < 0:
        # If we have options we haven't tried yet
        if len(_untried):
            # # tryNext = _untried.pop(0)
            # The next recursion will remove it for us
            tryNext = _untried[0]
            if _verbose:
                print(f'Not found using {backend}, trying {tryNext}')
            return invert(expr, words, randomNumbers, alot, _tried, tryNext, _verbose, _untried) # type: ignore
        # If we've tried everything, go to the error outside the if statement
    # If we're here, we generated a bad string, but we have more tries
    else:
        if _verbose:
            print(f'Not found using {backend}, retrying')

        return invert(expr, words, randomNumbers, alot, tries, backend, _verbose, _untried, _tried-1) # type: ignore

    raise NotImplementedError(f"Failed to invert pattern `{expr}`. Likely, bad regex was given. "
        "Otherwise, there\'s a bug in the invert function. You can submit a bug report to "
        "smartycope@gmail.com, and include the regex you used to get this error. (Failed expr: "
        f"was `{rtn}`)")
    return rtn


if __name__ == '__main__':
    pass
    # print(invert(r'D(?=AB)C'))  # ASSERT,      1, ifProcededBy
    # print(invert(r'D(?!AB)C'))  # ASSERT_NOT,  1, ifNotProcededBy
    # print(invert(r'C(?<=AB)D', _verbose=True)) # ASSERT,     -1, ifPrecededBy
    # print(invert(r'C(?<!AB)D'))   # ASSERT_NOT, -1, ifNotPrecededBy

    # print(invert(r'C(?=AB)'))  # ASSERT,      1, ifProcededBy
    # print(invert(r'C(?!AB)'))  # ASSERT_NOT,  1, ifNotProcededBy
    # print(invert(r'(?<=AB)C')) # ASSERT,     -1, ifPrecededBye
    # print(invert(r'(?<!AB)C'))   # ASSERT_NOT, -1, ifNotPrecededBy

    # print(invert(r'(?=AB)(?!CD)DC AB(?<=CD) AB(?<!CD)'))

    # print(invert(r'[ABC]+(?=D).*$ <.*?>'))

    # print(invert(r'\w+test\d+', _verbose=True))
